@@ -3,6 +3,8 @@
  * reiserfsprogs/README
  */
 
+#define _GNU_SOURCE
+
 #include "includes.h"
 #include <stdarg.h>
 #include <limits.h>
@@ -330,8 +332,7 @@ void print_directory_item (FILE * fp, reiserfs_filsys_t * fs,
 			  GET_HASH_VALUE (get_deh_offset (deh)),
 			  GET_GENERATION_NUMBER (get_deh_offset (deh)),
 			  get_deh_location (deh), get_deh_state (deh),
-			  code2name (find_hash_in_use (name, namelen, 
-						       GET_HASH_VALUE (get_deh_offset (deh)),
+			  code2name (find_hash_in_use (name, namelen, get_deh_offset (deh),
 						       fs ? get_sb_hash_code (fs->fs_ondisk_sb) : UNSET_HASH)));
 	/*fs ? (is_properly_hashed (fs, name, namelen, deh_offset (deh)) ? "" : "(BROKEN)") : "??");*/
     }
@@ -392,9 +393,9 @@ void print_indirect_item (FILE * fp, struct buffer_head * bh, int item_num)
     reiserfs_warning (fp, "%d pointer%s\n[", I_UNFM_NUM (ih),
                       I_UNFM_NUM (ih) != 1 ? "s" : "" );
     for (j = 0; j < I_UNFM_NUM (ih); j ++) {
-	if (sequence_finished (prev, &num, unp[j])) {
+	if (sequence_finished (prev, &num, d32_get(unp, j))) {
 	    print_sequence (fp, prev, num);
-	    start_new_sequence (&prev, &num, unp[j]);
+	    start_new_sequence (&prev, &num, d32_get(unp, j));
 	}
     }
     print_sequence (fp, prev, num);
@@ -617,6 +618,10 @@ int print_super_block (FILE * fp, reiserfs_filsys_t * fs, char * file_name,
 
     rdev = misc_device_rdev(file_name);
 
+    /* Print volume label if it is non-empty. */
+    if (sb->s_label[0])
+        reiserfs_warning (fp, "%s: ", sb->s_label);
+
     reiserfs_warning (fp, "Reiserfs super block in block %lu on 0x%x of ",
 		      bh->b_blocknr, rdev);
     switch (get_reiserfs_format (sb)) {
@@ -646,7 +651,7 @@ int print_super_block (FILE * fp, reiserfs_filsys_t * fs, char * file_name,
 		      "bitmaps, data, reserved] blocks): %u\n", get_sb_free_blocks (sb));
 	reiserfs_warning (fp, "Root block: %u\n", get_sb_root_block (sb));
     }
-    reiserfs_warning (fp, "Filesystem marked as %scleanly umounted\n",
+    reiserfs_warning (fp, "Filesystem is %sclean\n",
 		      (get_sb_umount_state (sb) == FS_CLEANLY_UMOUNTED) ? "" : "NOT ");
 
     if (short_print)
