@@ -1,14 +1,17 @@
 /*
- *  Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
+ *  Copyright 2000-2004 by Hans Reiser, licensing governed by 
+ *  reiserfsprogs/README
  */
 
+#define _GNU_SOURCE
+
 #include "includes.h"
-#include <linux/major.h>
 #include <linux/kdev_t.h>
 
 struct key root_dir_key = {0, 0, {{0, 0},}};
 struct key parent_root_dir_key = {0, 0, {{0, 0},}};
 struct key lost_found_dir_key = {0, 0, {{0, 0}, }};
+struct key badblock_key = {BADBLOCK_DIRID, BADBLOCK_OBJID, {{0, 0},}};
 
 __u16 root_dir_format = 0;
 __u16 lost_found_dir_format = 0;
@@ -27,13 +30,13 @@ static void make_const_keys (void)
 /* reiserfs needs at least: enough blocks for journal, 64 k at the beginning,
    one block for super block, bitmap block and root block. Note that first
    bitmap block must point to all of them */
-int is_block_count_correct (unsigned long block_of_super_block, int block_size,
-	unsigned long block_count, unsigned long journal_size)
+int is_block_count_correct (unsigned long journal_offset, unsigned int block_size,
+    unsigned long block_count, unsigned long journal_size)
 {
     unsigned long blocks;
 
-    /* RESERVED, MD RAID SBs, super block, bitmap, root, journal size w/ journal header */
-    blocks = block_of_super_block + 1 + 1 + 1 + journal_size;
+    /* RESERVED, MD RAID SBs, super block, bitmap, root, journal size with journal header */
+    blocks = journal_offset + journal_size;
 
     /* we have a limit: skipped area, super block, journal and root block
     all have to be addressed by one first bitmap */
@@ -47,20 +50,27 @@ int is_block_count_correct (unsigned long block_of_super_block, int block_size,
 }
 
 /* read super block. fixme: only 4k blocks, pre-journaled format
-   is refused. Journal and bitmap are to be opened separately */
-reiserfs_filsys_t * reiserfs_open (char * filename, int flags, int *error, void * vp)
+   is refused. Journal and bitmap are to be opened separately.
+   skip_check is set to 1 if checks of openned SB should be omitted.*/
+reiserfs_filsys_t * reiserfs_open (char * filename, int flags, 
+				   int *error, void * vp, int check)
 {
     reiserfs_filsys_t * fs;
     struct buffer_head * bh;
     struct reiserfs_super_block * sb;
     int fd, i;
 
-
     /* convert root dir key and parent root dir key to little endian format */
     make_const_keys ();
 
+    if (error) 
+	*error = 0;
 
-    fd = open (filename, flags | O_LARGEFILE);
+    fd = open (filename, flags 
+#if defined(O_LARGEFILE)
+	       | O_LARGEFILE
+#endif
+	       );
     if (fd == -1) {
 	if (error)
 	    *error = errno;
@@ -81,9 +91,7 @@ reiserfs_filsys_t * reiserfs_open (char * filename, int flags, int *error, void
 	} else {
 	    sb = (struct reiserfs_super_block *)bh->b_data;
 	    
-	    if (is_any_reiserfs_magic_string (sb))
-		/* fixme: we could make some check to make sure that super
-		   block looks correctly */
+	    if (is_any_reiserfs_magic_string(sb))
 		goto found;
 
 	    /* reiserfs signature is not found at the i-th 4k block */
@@ -91,19 +99,47 @@ reiserfs_filsys_t * reiserfs_open (char * filename, int flags, int *error, void
 	}
     }
 
-    reiserfs_warning (stderr, "reiserfs_open: neither new nor old reiserfs format "
-		      "found on %s\n", filename);
-    if (error)
-	*error = 0;
+    reiserfs_warning(stderr, 
+	"\nreiserfs_open: the reiserfs superblock cannot be found on %s.\n", filename);
+    
     freemem (fs);
     close (fd);
     fs = NULL;
     return fs;
 
  found:
-    
-    fs->fs_blocksize = get_sb_block_size (sb);
 
+    if (!is_blocksize_correct(get_sb_block_size(sb))) {
+	reiserfs_warning(stderr, "reiserfs_open: a superblock with wrong parameters "
+			 "was found in the block (%d).\n", i);
+	freemem (fs);
+	close (fd);
+	brelse(bh);
+	return NULL;
+    }
+
+    if (check) {
+	/* A few checks of found super block. */
+	struct buffer_head *tmp_bh;
+	
+	tmp_bh = bread (fd, get_sb_block_count(sb) - 1, get_sb_block_size(sb));
+	
+	if (!tmp_bh) {
+	    reiserfs_warning (stderr, "\n%s: Your partition is not big enough to contain the \n"
+		    "filesystem of (%lu) blocks as was specified in the found super block.\n", 
+		    __FUNCTION__,  get_sb_block_count(sb) - 1);
+	    
+	    freemem (fs);
+	    close (fd);
+	    brelse(bh);
+	    return NULL;
+	}
+	
+	brelse(tmp_bh);
+    }
+   
+    fs->fs_blocksize = get_sb_block_size (sb);
+    
     /* check block size on the filesystem */
     if (fs->fs_blocksize != 4096) {
 	i = bh->b_blocknr * 4096 / fs->fs_blocksize;
@@ -115,13 +151,7 @@ reiserfs_filsys_t * reiserfs_open (char * filename, int flags, int *error, void
 	    freemem (fs);
 	    return 0;
 	}
-	sb = (struct reiserfs_super_block *)bh->b_data;	
-	if (!does_look_like_super_block (sb, fs->fs_blocksize)) {
-	    reiserfs_warning (stderr, "reiserfs_open: no reiserfs in %d, size %d\n",
-			      i, fs->fs_blocksize);
-	    freemem (fs);
-	    return 0;
-	}
+	sb = (struct reiserfs_super_block *)bh->b_data;
     }
 
     fs->fs_hash_function = code2func (get_sb_hash_code (sb));
@@ -145,7 +175,10 @@ reiserfs_filsys_t * reiserfs_open (char * filename, int flags, int *error, void
    constant for given size and version of a filesystem */
 reiserfs_filsys_t * reiserfs_create (char * filename,
 				     int version,
-				     unsigned long block_count, int block_size, int default_journal, int new_format)
+				     unsigned long block_count, 
+				     int block_size, 
+				     int default_journal, 
+				     int new_format)
 {
     reiserfs_filsys_t * fs;
 
@@ -159,11 +192,11 @@ reiserfs_filsys_t * reiserfs_create (char * filename,
 	return 0;
     }
 
-    if (!is_block_count_correct (REISERFS_DISK_OFFSET_IN_BYTES / block_size, block_size, block_count, 0)) {
-//    if (block_count < min_block_amount (block_size, 0)) {
-	reiserfs_warning (stderr, "reiserfs_create: "
-			  "can not create that small (%d blocks) filesystem\n",
-			  block_count);
+    if (!is_block_count_correct (REISERFS_DISK_OFFSET_IN_BYTES / block_size, 
+	block_size, block_count, 0)) 
+    {
+	reiserfs_warning (stderr, "reiserfs_create: can not create that small "
+	    "(%d blocks) filesystem\n", block_count);
 	return 0;
     }
 
@@ -173,10 +206,14 @@ reiserfs_filsys_t * reiserfs_create (char * filename,
 	return 0;
     }
 
-    fs->fs_dev = open (filename, O_RDWR | O_LARGEFILE);
+    fs->fs_dev = open (filename, O_RDWR 
+#if defined(O_LARGEFILE)
+		       | O_LARGEFILE
+#endif
+		       );
     if (fs->fs_dev == -1) {
-	reiserfs_warning (stderr, "reiserfs_create: could not open %s: %m\n",
-			  filename);
+	reiserfs_warning (stderr, "reiserfs_create: could not open %s: %s\n",
+			  filename, strerror(errno));
 	freemem (fs);
 	return 0;
     }
@@ -186,9 +223,11 @@ reiserfs_filsys_t * reiserfs_create (char * filename,
     fs->fs_format = version;
 
     if (new_format)
-        fs->fs_super_bh = getblk (fs->fs_dev, REISERFS_DISK_OFFSET_IN_BYTES / block_size, block_size);
+        fs->fs_super_bh = getblk (fs->fs_dev, 
+	    REISERFS_DISK_OFFSET_IN_BYTES / block_size, block_size);
     else 
-        fs->fs_super_bh = getblk (fs->fs_dev, REISERFS_OLD_DISK_OFFSET_IN_BYTES / block_size, block_size);
+        fs->fs_super_bh = getblk (fs->fs_dev, 
+	    REISERFS_OLD_DISK_OFFSET_IN_BYTES / block_size, block_size);
     
     if (!fs->fs_super_bh) {
 	reiserfs_warning (stderr, "reiserfs_create: getblk failed\n");
@@ -215,7 +254,8 @@ reiserfs_filsys_t * reiserfs_create (char * filename,
     set_sb_block_size (fs->fs_ondisk_sb, block_size);
     switch (version) {
     case REISERFS_FORMAT_3_5:
-	set_sb_oid_maxsize (fs->fs_ondisk_sb, (block_size - SB_SIZE_V1) / sizeof(__u32) / 2 * 2);
+	set_sb_oid_maxsize (fs->fs_ondisk_sb, 
+	    (block_size - SB_SIZE_V1) / sizeof(__u32) / 2 * 2);
 	/* sb_oid_cursize */
 	/* sb_state */
 	memcpy (fs->fs_ondisk_sb->s_v1.s_magic, REISERFS_3_5_SUPER_MAGIC_STRING,
@@ -223,7 +263,8 @@ reiserfs_filsys_t * reiserfs_create (char * filename,
 	break;
 
     case REISERFS_FORMAT_3_6:
-	set_sb_oid_maxsize (fs->fs_ondisk_sb, (block_size - SB_SIZE) / sizeof(__u32) / 2 * 2);
+	set_sb_oid_maxsize (fs->fs_ondisk_sb, 
+	    (block_size - SB_SIZE) / sizeof(__u32) / 2 * 2);
 	/* sb_oid_cursize */
 	/* sb_state */
         memcpy (fs->fs_ondisk_sb->s_v1.s_magic, REISERFS_3_6_SUPER_MAGIC_STRING,
@@ -270,24 +311,15 @@ int spread_bitmaps (reiserfs_filsys_t * fs)
 /* 0 means: do not guarantee that fs is consistent */
 int reiserfs_is_fs_consistent (reiserfs_filsys_t * fs)
 {
-    if (get_sb_umount_state (fs->fs_ondisk_sb) == REISERFS_CLEANLY_UMOUNTED &&
-	get_sb_fs_state (fs->fs_ondisk_sb) == REISERFS_CONSISTENT)
+    if (get_sb_umount_state (fs->fs_ondisk_sb) == FS_CLEANLY_UMOUNTED &&
+	get_sb_fs_state (fs->fs_ondisk_sb) == FS_CONSISTENT)
 	return 1;
     return 0;
 }
 
-
-void reiserfs_reopen (reiserfs_filsys_t * fs, int flag)
-{
-    reiserfs_only_reopen (fs, flag);
-    reiserfs_reopen_journal (fs, flag);
-
-}
-
-
 /* flush bitmap, brelse super block, flush all dirty buffers, close and open
    again the device, read super block */
-void reiserfs_only_reopen (reiserfs_filsys_t * fs, int flag)
+static void reiserfs_only_reopen (reiserfs_filsys_t * fs, int flag)
 {
     unsigned long super_block;
 
@@ -299,11 +331,15 @@ void reiserfs_only_reopen (reiserfs_filsys_t * fs, int flag)
     
     invalidate_buffers (fs->fs_dev);
     if (close (fs->fs_dev))
-	die ("reiserfs_reopen: closed failed: %m");
+	die ("reiserfs_reopen: closed failed: %s", strerror(errno));
     
-    fs->fs_dev = open (fs->fs_file_name, flag | O_LARGEFILE);
+    fs->fs_dev = open (fs->fs_file_name, flag 
+#if defined(O_LARGEFILE)
+		       | O_LARGEFILE
+#endif
+		       );
     if (fs->fs_dev == -1)
-	die ("reiserfs_reopen: could not reopen device: %m");
+	die ("reiserfs_reopen: could not reopen device: %s", strerror(errno));
 
     fs->fs_super_bh = bread (fs->fs_dev, super_block, fs->fs_blocksize);
     if (!fs->fs_super_bh)
@@ -317,6 +353,11 @@ void reiserfs_only_reopen (reiserfs_filsys_t * fs, int flag)
 	fs->fs_dirt = 0;
 }
 
+void reiserfs_reopen (reiserfs_filsys_t * fs, int flag)
+{
+    reiserfs_only_reopen (fs, flag);
+    reiserfs_reopen_journal (fs, flag);
+}
 
 int is_opened_rw (reiserfs_filsys_t * fs)
 {
@@ -368,7 +409,9 @@ void reiserfs_close (reiserfs_filsys_t * fs)
 
 
 int reiserfs_new_blocknrs (reiserfs_filsys_t * fs, 
-			   unsigned long * free_blocknrs, unsigned long start, int amount_needed)
+			   unsigned long * free_blocknrs, 
+			   unsigned long start, 
+			   int amount_needed)
 {
     if (fs->block_allocator)
 	return fs->block_allocator (fs, free_blocknrs, start, amount_needed);
@@ -385,18 +428,6 @@ int reiserfs_free_block (reiserfs_filsys_t * fs, unsigned long block)
     return 0;
 }
 
-/*
-inline int reiserfs_bin_search (void * key, void * base, int num, int width,
-			 int *ppos, comparison_fn_t comp_func) {
-    __u32 position = *ppos;
-    int retval;
-
-    retval = reiserfs_bin_search (key, base, num, width, &position, comp_func);
-    *ppos = position;
-    return retval;
-}
-*/
-
 static int reiserfs_search_by_key_x (reiserfs_filsys_t * fs, struct key * key,
 				     struct path * path, int key_length)
 {
@@ -407,6 +438,9 @@ static int reiserfs_search_by_key_x (reiserfs_filsys_t * fs, struct key * key,
     
 
     block = get_sb_root_block (fs->fs_ondisk_sb);
+    if (not_data_block(fs, block))
+	return IO_ERROR;
+    
     path->path_length = ILLEGAL_PATH_ELEMENT_OFFSET;
     while (1) {
 	curr = PATH_OFFSET_PELEMENT (path, ++ path->path_length);
@@ -432,6 +466,8 @@ static int reiserfs_search_by_key_x (reiserfs_filsys_t * fs, struct key * key,
 		return ITEM_NOT_FOUND;
 	}
 	block = get_dc_child_blocknr (B_N_CHILD (bh, curr->pe_position));
+	if (not_data_block(fs, block))
+		return IO_ERROR;
     }
     printf ("search_by_key: you can not get here\n");
     return ITEM_NOT_FOUND;
@@ -539,16 +575,14 @@ int usearch_by_position (reiserfs_filsys_t * s, struct key * key,
 
 static int comp_dir_entries (const void * p1, const void * p2)
 {
-    __u32 deh_offset;
-    const __u32 * off1, * off2;
+    __u32 off1, off2;
 
-    off1 = p1;
-    off2 = p2;
-    deh_offset = le32_to_cpu (*off1);
-
-    if (deh_offset < *off2)
+    off1 = d32_get((__u32 *)p1, 0);
+    off2 = *(__u32 *)p2;
+    
+    if (off1 < off2)
 	return -1;
-    if (deh_offset > *off2)
+    if (off1 > off2)
 	return 1;
     return 0;
 }
@@ -623,15 +657,7 @@ struct key * uget_rkey (struct path * path)
     return 0;
 }
 
-struct key * get_next_key_2 (struct path * path)
-{
-    if (PATH_LAST_POSITION (path) < B_NR_ITEMS (get_bh (path)) - 1)
-	return B_N_PKEY (get_bh (path), PATH_LAST_POSITION (path) + 1);
-    return uget_rkey (path);
-}
-
-struct key * reiserfs_next_key (struct path * path)
-{
+struct key * reiserfs_next_key (struct path * path) {
     if (get_item_pos (path) < B_NR_ITEMS (get_bh (path)) - 1)
 	return B_N_PKEY (get_bh (path), get_item_pos (path) + 1);
 
@@ -813,16 +839,11 @@ void reiserfs_insert_item (reiserfs_filsys_t * fs, struct path * path,
 
 /*===========================================================================*/
 
-static __u32 hash_value (reiserfs_filsys_t * fs, char * name)
+__u32 hash_value (hashf_t func, char * name, int namelen)
 {
     __u32 res;
 
-    if (!strcmp (name, "."))
-	return DOT_OFFSET;
-    if (!strcmp (name, ".."))
-	return DOT_DOT_OFFSET;
-
-    res = reiserfs_hash (fs) (name, strlen (name));    
+    res = func (name, namelen);
     res = GET_HASH_VALUE(res);
     if (res == 0)
 	res = 128;
@@ -839,14 +860,12 @@ int reiserfs_locate_entry (reiserfs_filsys_t * fs, struct key * dir, char * name
     struct key entry_key;
     struct item_head * ih;
     struct reiserfs_de_head * deh;
-    __u32 hash;
     int i, retval;
     struct key * rdkey;
     
     set_key_dirid (&entry_key, get_key_dirid (dir));
     set_key_objectid (&entry_key, get_key_objectid (dir));
-    hash = hash_value (fs, name);
-    set_key_offset_v1 (&entry_key, hash);
+    set_key_offset_v1 (&entry_key, 0);
     set_key_uniqueness (&entry_key, DIRENTRY_UNIQUENESS);
 
  
@@ -859,14 +878,8 @@ int reiserfs_locate_entry (reiserfs_filsys_t * fs, struct key * dir, char * name
 	ih = get_ih (path);
 	deh = B_I_DEH (get_bh (path), ih) + path->pos_in_item;
 	for (i = path->pos_in_item; i < get_ih_entry_count (ih); i ++, deh ++) {
-	    if (GET_HASH_VALUE (get_deh_offset (deh)) != GET_HASH_VALUE (hash)) {
-		/* all entries having the same hash were scanned */
-		pathrelse (path);
-		return 0;
-	    }
-
 	    /* the name in directory has the same hash as the given name */
-	    if ((name_in_entry_length (ih, deh, i) == strlen (name)) &&
+	    if ((name_in_entry_length (ih, deh, i) == (int)strlen (name)) &&
 		!memcmp (name_in_entry (deh, i), name, strlen (name))) {
 		path->pos_in_item = i;
 		return 1;
@@ -882,13 +895,6 @@ int reiserfs_locate_entry (reiserfs_filsys_t * fs, struct key * dir, char * name
 	if (!is_direntry_key (rdkey))
 	    reiserfs_panic ("reiserfs_locate_entry: can not find name in broken directory yet");
 
-	/* next item is the item of the directory we are looking name in */
-	if (GET_HASH_VALUE (get_offset (rdkey)) != hash) {
-	    /* but there is no names with given hash */
-	    pathrelse (path);
-	    return 0;
-	}
-
 	/* first name of that item may be a name we are looking for */
 	entry_key = *rdkey;
 	pathrelse (path);
@@ -908,7 +914,7 @@ int reiserfs_locate_entry (reiserfs_filsys_t * fs, struct key * dir, char * name
    generation counter in 'min_gen_counter'. dies if found object is not a
    directory. */
 int reiserfs_find_entry (reiserfs_filsys_t * fs, struct key * dir, char * name, 
-			 int * min_gen_counter, struct key * key)
+			 unsigned int * min_gen_counter, struct key * key)
 {
     struct key entry_key;
     int retval;
@@ -922,7 +928,12 @@ int reiserfs_find_entry (reiserfs_filsys_t * fs, struct key * dir, char * name,
 
     set_key_dirid (&entry_key, get_key_dirid (dir));
     set_key_objectid (&entry_key, get_key_objectid (dir));
-    hash = hash_value (fs, name);
+    if (!strcmp (name, "."))
+	hash = DOT_OFFSET;
+    else if (!strcmp (name, ".."))
+	hash = DOT_DOT_OFFSET;
+    else
+	hash = hash_value (reiserfs_hash (fs), name, strlen (name));
     set_key_offset_v1 (&entry_key, hash);
     set_key_uniqueness (&entry_key, DIRENTRY_UNIQUENESS);
 
@@ -946,7 +957,7 @@ int reiserfs_find_entry (reiserfs_filsys_t * fs, struct key * dir, char * name,
 	    if (GET_GENERATION_NUMBER (get_deh_offset (deh)) == *min_gen_counter)
 		(*min_gen_counter) ++;
 			
-	    if ((name_in_entry_length (ih, deh, i) == strlen (name)) &&
+	    if ((name_in_entry_length (ih, deh, i) == (int)strlen (name)) &&
 	        (!memcmp (name_in_entry (deh, i), name, strlen (name)))) {
 		/* entry found in the directory */
 		if (key) {
@@ -1017,13 +1028,13 @@ char * make_entry (char * entry, char * name, struct key * key, __u32 offset)
 /* add new name into a directory. If it exists in a directory - do
    nothing */
 int reiserfs_add_entry (reiserfs_filsys_t * fs, struct key * dir, char * name, int name_len,
-			struct key * key, int fsck_need)
+			struct key * key, __u16 fsck_need)
 {
     struct item_head entry_ih = {{0,}, };
     char * entry;
     int retval;
     INITIALIZE_PATH(path);
-    int gen_counter;
+    unsigned int gen_counter;
     int item_len;
     __u32 hash;
 
@@ -1034,11 +1045,12 @@ int reiserfs_add_entry (reiserfs_filsys_t * fs, struct key * dir, char * name, i
     /* compose entry key to look for its place in the tree */
     set_key_dirid (&(entry_ih.ih_key), get_key_dirid (dir));
     set_key_objectid (&(entry_ih.ih_key), get_key_objectid (dir));
-    hash = hash_value (fs, name) + gen_counter;
     if (!strcmp (name, "."))
 	hash = DOT_OFFSET;
-    if (!strcmp (name, ".."))
+    else if (!strcmp (name, ".."))
 	hash = DOT_DOT_OFFSET;
+    else
+	hash = hash_value (reiserfs_hash (fs), name, strlen (name)) + gen_counter;
     set_key_offset_v1 (&(entry_ih.ih_key), hash);
     set_key_uniqueness (&(entry_ih.ih_key), DIRENTRY_UNIQUENESS);
 
@@ -1116,6 +1128,7 @@ int create_dir_sd (reiserfs_filsys_t * fs,
     else
 	key_format = KEY_FORMAT_2;
 
+    memset(&sd, 0, sizeof(sd));
     make_dir_stat_data (fs->fs_blocksize, key_format, get_key_dirid (key),
 			get_key_objectid (key), &ih, &sd);
 
@@ -1171,50 +1184,24 @@ int can_we_format_it (char * device_name, int force)
     dev_t rdev;
 
 
-    if (is_mounted (device_name)) {
+    if (misc_device_mounted(device_name) > 0) {
 	/* device looks mounted */
 	reiserfs_warning (stderr, "'%s' looks mounted.", device_name);
 	check_forcing_ask_confirmation (force);
     }
 
-    mode = get_st_mode (device_name);
-    rdev = get_st_rdev (device_name);
+    mode = misc_device_mode(device_name);
+    rdev = misc_device_rdev(device_name);
 
     if (!S_ISBLK (mode)) {
 	/* file is not a block device */
-	reiserfs_warning (stderr, "%s is not a block special device", device_name);
+	reiserfs_warning (stderr, "%s is not a block special device\n", device_name);
 	check_forcing_ask_confirmation (force);
     } else {
-	/* from e2progs-1.18/misc/mke2fs.c */
-
-	
-/*FIXME: it should be probably rewritten for 2.5 later*/
-#ifndef MAJOR
-#define MAJOR(dev)	((dev)>>8)
-#define MINOR(dev)	((dev) & 0xff)
-#endif
-#ifndef SCSI_BLK_MAJOR
-#define SCSI_BLK_MAJOR(M)  ((M) == SCSI_DISK_MAJOR || (M) == SCSI_CDROM_MAJOR)
-#endif
-
-#ifndef IDE_DISK_MAJOR
-#ifdef IDE9_MAJOR
-#define IDE_DISK_MAJOR(M) (MAJOR(rdev) == IDE0_MAJOR || MAJOR(rdev) == IDE1_MAJOR || \
-			   MAJOR(rdev) == IDE2_MAJOR || MAJOR(rdev) == IDE3_MAJOR || \
-			   MAJOR(rdev) == IDE4_MAJOR || MAJOR(rdev) == IDE5_MAJOR || \
-			   MAJOR(rdev) == IDE6_MAJOR || MAJOR(rdev) == IDE7_MAJOR || \
-			   MAJOR(rdev) == IDE8_MAJOR || MAJOR(rdev) == IDE9_MAJOR)
-#else
-#define IDE_DISK_MAJOR(M) (MAJOR(rdev) == IDE0_MAJOR || MAJOR(rdev) == IDE1_MAJOR || \
-			   MAJOR(rdev) == IDE2_MAJOR || MAJOR(rdev) == IDE3_MAJOR || \
-			   MAJOR(rdev) == IDE4_MAJOR || MAJOR(rdev) == IDE5_MAJOR)
-#endif
-#endif
-
-	if ((IDE_DISK_MAJOR (MAJOR(rdev)) && MINOR (rdev) % 64 == 0) ||
-	    (SCSI_BLK_MAJOR (MAJOR(rdev)) && MINOR (rdev) % 16 == 0)) {
+	if ((IDE_DISK_MAJOR (major(rdev)) && minor(rdev) % 64 == 0) ||
+	    (SCSI_BLK_MAJOR (major(rdev)) && minor(rdev) % 16 == 0)) {
 	    /* /dev/hda or similar */
-	    reiserfs_warning (stderr, "%s is entire device, not just one partition!",
+	    reiserfs_warning (stderr, "%s is entire device, not just one partition!\n",
 		    device_name);
 	    check_forcing_ask_confirmation (force);
 	}
@@ -1223,123 +1210,26 @@ int can_we_format_it (char * device_name, int force)
     return 1;
 }
 
-	
-#if 0	
-/*tune*/
-#define check_forcing_ask_confirmation() \
-	if (!Force) {\
-	    /* avoid formatting it without being forced */\
-	    reiserfs_warning (stderr, "Use -f to force over\n");\
-	    return 0;\
-	}\
-	if (Force == 1) {\
-	    if (!user_confirmed (stderr, "Continue (y/n):", "y\n"))\
-		return 0;\
-	}\
-
-
-/* we only can use a file for filesystem or journal if it is either not
-   mounted block device or regular file and we are forced to use it */
-static int can_we_format_it (char * device_name)
-{
-    mode_t mode;
-    dev_t rdev;
-
-
-    if (is_mounted (device_name)) {
-	/* device looks mounted */
-	message("'%s' looks mounted.\n", device_name);
-	check_forcing_ask_confirmation ();
-    }
-
-    mode = get_st_mode (device_name);
-    rdev = get_st_rdev (device_name);
-
-    if (!S_ISBLK (mode)) {
-	/* file is not a block device */
-	message("Can not create a journal on not a block device file %s\n", device_name);
-	exit (1);
-    } else {
-	/* from e2progs-1.18/misc/mke2fs.c */
-	if ((MAJOR (rdev) == HD_MAJOR && MINOR (rdev) % 64 == 0) ||
-	    (SCSI_BLK_MAJOR (rdev) && MINOR (rdev) % 16 == 0)) {
-	    /* /dev/hda or similar */
-	    message("%s is entire device, not just one partition!\n",
-		    device_name);
-	    check_forcing_ask_confirmation ();
-	}
-    }
-
-    return 1;
-}
-
-
-/*mkreiserfs*/
-/* we only can use a file for filesystem or journal if it is either not
-   mounted block device or regular file and we are forced to use it */
-static int can_we_format_it (char * device_name, int force)
-{
-    mode_t mode;
-    dev_t rdev;
-
-    if (is_mounted (device_name)) {
-		/* device looks mounted */
-	    message("'%s' looks mounted.", device_name);
-		check_forcing_ask_confirmation ();
-    }
-
-    mode = get_st_mode (device_name);
-    rdev = get_st_rdev (device_name);
-
-    if (!S_ISBLK (mode)) {
-		/* file is not a block device */
-	    message("%s is not a block special device", device_name);
-		check_forcing_ask_confirmation ();
-    } else {
-		if (((MAJOR (rdev) == IDE0_MAJOR ||
-			  MAJOR (rdev) == IDE1_MAJOR ||
-			  MAJOR (rdev) == IDE2_MAJOR ||
-			  MAJOR (rdev) == IDE3_MAJOR ||
-			  MAJOR (rdev) == IDE4_MAJOR ||
-			  MAJOR (rdev) == IDE5_MAJOR ||
-			  MAJOR (rdev) == IDE6_MAJOR ||
-			  MAJOR (rdev) == IDE7_MAJOR ||
-			  MAJOR (rdev) == IDE8_MAJOR ||
-			  MAJOR (rdev) == IDE9_MAJOR) &&
-			 MINOR (rdev) % 64 == 0) ||
-			(SCSI_BLK_MAJOR (MAJOR (rdev)) &&
-			 MINOR (rdev) % 16 == 0)) {
-			/* /dev/hda or similar */
-			message("%s is entire device, not just one partition!",
-					device_name);
-			check_forcing_ask_confirmation ();
-		}
-    }
-
-    return 1;
-}
-	
-#endif
-	
-	
-
-
 int create_badblock_bitmap (reiserfs_filsys_t * fs, char * badblocks_file) {
     FILE * fd;
     char buf[128];
     __u32 blocknr;
     int count;
 
+    fs->fs_badblocks_bm = reiserfs_create_bitmap(get_sb_block_count(fs->fs_ondisk_sb));
+    reiserfs_bitmap_zero (fs->fs_badblocks_bm);
+
+    if (!badblocks_file)
+	return 0;
+    
     fd = fopen (badblocks_file, "r");
 
     if (fd == NULL) {
-        fprintf (stderr, "%s: could not open badblock file %s, work without it\n",
+        fprintf (stderr, "%s: Failed to open the given badblock file '%s'.\n\n",
         	__FUNCTION__, badblocks_file);
         return 1;
     }
 
-    fs->fs_badblocks_bm = reiserfs_create_bitmap (get_sb_block_count (fs->fs_ondisk_sb));
-    reiserfs_bitmap_zero (fs->fs_badblocks_bm);
 
     while (!feof (fd)) {
 	if (fgets(buf, sizeof(buf), fd) == NULL)
@@ -1349,11 +1239,18 @@ int create_badblock_bitmap (reiserfs_filsys_t * fs, char * badblocks_file) {
 	if (count <= 0)
 	    continue;
 
-	if (blocknr < get_sb_block_count (fs->fs_ondisk_sb) && !not_data_block (fs, blocknr)) {
-	    reiserfs_bitmap_set_bit (fs->fs_badblocks_bm, blocknr);
+	if (blocknr >= get_sb_block_count (fs->fs_ondisk_sb)) {
+	    fprintf (stderr, "%s: block number (%u) points out of fs size "
+		     "(%u).\n", __FUNCTION__, blocknr, 
+		     get_sb_block_count(fs->fs_ondisk_sb));
+	} else if (not_data_block (fs, blocknr)) {
+	    fprintf (stderr, "%s: block number (%u) belongs to system "
+		     "reiserfs area. It cannot be relocated.\n", 
+		     __FUNCTION__, blocknr);
+	    return 1;
 	} else {
-	    fprintf (stderr, "%s: bad block number %u in badblocks file\n", __FUNCTION__, blocknr);
-	}
+	    reiserfs_bitmap_set_bit (fs->fs_badblocks_bm, blocknr);
+	} 
     }
 
     fclose (fd);
@@ -1361,14 +1258,99 @@ int create_badblock_bitmap (reiserfs_filsys_t * fs, char * badblocks_file) {
     return 0;
 }
 
-void add_badblock_list (reiserfs_filsys_t * fs, int no_badblock_in_tree_yet) {
+void badblock_list(reiserfs_filsys_t * fs, badblock_func_t action, void *data) {
+    struct path badblock_path;
+    struct key rd_key = badblock_key;
+    struct key *key;
+    
+    badblock_path.path_length = ILLEGAL_PATH_ELEMENT_OFFSET;
+    set_type_and_offset (KEY_FORMAT_2, &badblock_key, 1, TYPE_INDIRECT);
+    
+    while (1) {
+	    if (reiserfs_search_by_key_4 (fs, &rd_key, &badblock_path) == IO_ERROR) {
+		fprintf (stderr, "%s: Some problems while searching by the key "
+			 "occured. Probably due to tree corruptions.\n",
+			 __FUNCTION__);
+		pathrelse (&badblock_path);
+		break;
+	    }
+		    
+
+	    if (get_blkh_nr_items (B_BLK_HEAD (get_bh (&badblock_path))) <= 
+		PATH_LAST_POSITION (&badblock_path)) 
+	    {
+		pathrelse (&badblock_path);
+		break;
+	    }
+
+	    rd_key = get_ih(&badblock_path)->ih_key;
+
+	    if (get_key_dirid(&rd_key) != BADBLOCK_DIRID || 
+		get_key_objectid(&rd_key) != BADBLOCK_OBJID ||
+		!KEY_IS_INDIRECT_KEY(&rd_key)) 
+	    {
+		pathrelse (&badblock_path);
+		break;
+	    }
+	
+	    if ((key = reiserfs_next_key(&badblock_path)))
+		rd_key = *key;
+	    else
+	        memset(&rd_key, 0, sizeof(rd_key));
+	    
+	    action(fs, &badblock_path, data);
+
+	    if (get_key_dirid(&rd_key) == 0)
+		break;
+    }
+}
+
+static void callback_badblock_rm(reiserfs_filsys_t *fs,
+				 struct path *badblock_path, 
+				 void *data) 
+{
+	struct tree_balance tb;
+	struct item_head * tmp_ih;
+
+	tmp_ih = get_ih(badblock_path);
+	memset (get_item (badblock_path), 0, get_ih_item_len (tmp_ih));
+
+	init_tb_struct (&tb, fs, badblock_path, 
+			-(IH_SIZE + get_ih_item_len(PATH_PITEM_HEAD(badblock_path))));
+
+	if (fix_nodes (M_DELETE, &tb, 0) != CARRY_ON)
+		die ("%s: fix_nodes failed", __FUNCTION__);
+
+	do_balance (/*tb.transaction_handle,*/ &tb, 0, 0, M_DELETE, 0/*zero num*/);
+}
+
+void mark_badblock(reiserfs_filsys_t *fs, 
+		   struct path *badblock_path, 
+		   void *data) 
+{
+	struct item_head *tmp_ih;
+	__u32 *ind_item;
+	__u32 i;
+
+	if (!fs->fs_badblocks_bm)
+	    create_badblock_bitmap(fs, NULL);
+	
+	tmp_ih = get_ih(badblock_path);
+	ind_item = (__u32 *)get_item(badblock_path);
+
+	for (i = 0; i < I_UNFM_NUM(tmp_ih); i++) {
+		reiserfs_bitmap_set_bit(fs->fs_badblocks_bm, 
+					d32_get(ind_item, i));
+	}
+	
+	pathrelse (badblock_path);
+}
+
+void add_badblock_list (reiserfs_filsys_t * fs, int replace) {
     struct tree_balance tb;
-    struct key badblock_key = {-1, -1, {{0, 0}}};
     struct path badblock_path;
-    struct item_head * tmp_ih;
     struct item_head badblock_ih;
-//    char item[UNFM_P_SIZE];
-    struct unfm_nodeinfo ni;
+    __u32 ni;
 
     __u64 offset;
     __u32 i, j;
@@ -1377,46 +1359,19 @@ void add_badblock_list (reiserfs_filsys_t * fs, int no_badblock_in_tree_yet) {
     	return;
 
     /* delete all items with badblock_key */
-    while (1) {
-	reiserfs_search_by_key_4 (fs, &badblock_key, &badblock_path);
-	
-	if (get_blkh_nr_items ( B_BLK_HEAD (get_bh(&badblock_path))) <= PATH_LAST_POSITION (&badblock_path)) {
-	    pathrelse (&badblock_path);
-            break;
-	}
-
-	tmp_ih = get_ih(&badblock_path);
-	
-	if (get_key_dirid(&tmp_ih->ih_key) != (__u32)-1 ||
-	    get_key_objectid(&tmp_ih->ih_key) != (__u32)-1 || !is_indirect_ih (tmp_ih)) {
-	    pathrelse (&badblock_path);
-	    break;
-	}
-
-	if (no_badblock_in_tree_yet)
-	    reiserfs_panic ("$s: bad block list found in the tree\n", __FUNCTION__);	
-	
-	memset (get_item (&badblock_path), 0, get_ih_item_len (tmp_ih));
-	
-	
-	init_tb_struct (&tb, fs, &badblock_path, -(IH_SIZE + get_ih_item_len(PATH_PITEM_HEAD (&badblock_path))));
-	
-	if (fix_nodes (M_DELETE, &tb, 0) != CARRY_ON)
-	    die ("%s: fix_nodes failed", __FUNCTION__);
-	
-	do_balance (/*tb.transaction_handle,*/ &tb, 0, 0, M_DELETE, 0/*zero num*/);
-    }
+    if (replace)
+	badblock_list(fs, callback_badblock_rm, NULL);
 
+    memset(&badblock_ih, 0, sizeof(badblock_ih));
     set_ih_key_format (&badblock_ih, KEY_FORMAT_2);
     set_ih_item_len (&badblock_ih, UNFM_P_SIZE);
     set_ih_free_space (&badblock_ih, 0);
     set_ih_location (&badblock_ih, 0);
-    set_key_dirid (&badblock_ih.ih_key, -1);
-    set_key_objectid (&badblock_ih.ih_key, -1);
+    set_key_dirid (&badblock_ih.ih_key, BADBLOCK_DIRID);
+    set_key_objectid (&badblock_ih.ih_key, BADBLOCK_OBJID);
     set_type (KEY_FORMAT_2, &badblock_ih.ih_key, TYPE_INDIRECT);
 
     j = 0;
-    ni.unfm_freespace = 0;
 
     /* insert all badblock pointers */
     for (i = 0; i < fs->fs_badblocks_bm->bm_bit_size; i++) {
@@ -1427,23 +1382,27 @@ void add_badblock_list (reiserfs_filsys_t * fs, int no_badblock_in_tree_yet) {
 	
 	offset = j * fs->fs_blocksize + 1;
 	set_offset (KEY_FORMAT_2, &badblock_ih.ih_key, offset);
-	ni.unfm_nodenum = cpu_to_le32 (i);
+	ni = cpu_to_le32 (i);
 	
-	retval = usearch_by_position (fs, &badblock_ih.ih_key, key_format (&badblock_ih.ih_key), &badblock_path);
+	retval = usearch_by_position (fs, &badblock_ih.ih_key, 
+				      key_format(&badblock_ih.ih_key), 
+				      &badblock_path);
 
 	switch (retval) {
 	case (FILE_NOT_FOUND):
-		init_tb_struct (&tb, fs, &badblock_path, IH_SIZE + get_ih_item_len(&badblock_ih));
-		if (fix_nodes (/*tb.transaction_handle,*/ M_INSERT, &tb, &badblock_ih/*, body*/) != CARRY_ON)
-		    die ("reiserfsck_insert_item: fix_nodes failed");
-		do_balance (/*tb.transaction_handle,*/ &tb, &badblock_ih, (void *)&ni.unfm_nodenum , M_INSERT, 0/*zero num*/);
-
-//		reiserfsck_insert_item (&badblock_path, &badblock_ih, item);
+		init_tb_struct (&tb, fs, &badblock_path, 
+				IH_SIZE + get_ih_item_len(&badblock_ih));
 		
+		if (fix_nodes (M_INSERT, &tb, &badblock_ih) != CARRY_ON)
+		    die ("reiserfsck_insert_item: fix_nodes failed");
+		
+		do_balance (&tb, &badblock_ih, (void *)&ni , M_INSERT, 0);
+
 		break;
 	
 	case (POSITION_NOT_FOUND):
-		/* take unformatted pointer from an indirect item */
+	case (POSITION_FOUND):
+		/* Insert the new item to the found position. */
 	
 		init_tb_struct (&tb, fs, &badblock_path, UNFM_P_SIZE);
 		
@@ -1451,10 +1410,6 @@ void add_badblock_list (reiserfs_filsys_t * fs, int no_badblock_in_tree_yet) {
 		    die ("reiserfsck_paste_into_item: fix_nodes failed");
 
 		do_balance (&tb, 0, (const char *)&ni, M_PASTE, 0);
-						
-//		reiserfsck_file_write (&badblock_ih, item, 1 /* was in tree to avoid problems with bitmaps */);
-//		reiserfsck_paste_into_item (path, (const char *)&ni, UNFM_P_SIZE);
-		
 		break;
 	}
 	
